import type NetV from "netv/src"
import type { LinkData, LinkStyle, NodeData, NodeStyle } from "netv/src/interfaces"

export interface DisplayNodeData extends NodeData {
  /** (Required) x-coordinate in range [0, 1]. */
  x: number,
  /** (Required) y-coordinate in range [0, 1]. */
  y: number,
  /** 
   * Node ID in the next level to which current node is coarsened.
   * Not present in the final level.
   */
  to?: string,
}

export interface DisplayLevelJSONData {
  /** All the nodes in the current level. */
  nodes: Array<DisplayNodeData>,
  /** All the links in the current level. */
  links: Array<LinkData>
}

export interface DisplayLevelData {
  /** All the nodes in the current level. Finds node by ID. */
  nodes: Map<string, DisplayNodeData>,
  /** All the links in the current level. */
  links: Array<LinkData>
}

export type IMAlgorithmKey = 'cwim' | 'contrast';

interface DisplayDataBase<
    LevelData extends DisplayLevelJSONData | DisplayLevelData, 
    SeedContainer extends Array<string> | Set<string>,
> {
  /** Graphs in each level. */
  graphs: Array<LevelData>,
  /** Seed nodes from different algorithms. */
  seeds: {
    /** 
     * Seed nodes obtained by C-w-IM algorithm in each level.
     * It shall be guaranteed in prior that each seeds in the (i + 1)-th layer
     * is coarsened from some seed in the i-th layer.
     */
    cwim: Array<SeedContainer>,
    /** Seeds nodes obtained by the contrast IM algorithm. */
    contrast: SeedContainer,
  },
  /** 
   * Process of information propagation from seeds by discrete steps.
   * simulation[t] represents propagation actions (u -> v) in the t-th iteration. 
   */
  simulation: {
    cwim: Array<Array<LinkData>>,
    contrast: Array<Array<LinkData>>,
  },
  /** Time usage of different algorithms */
  timeUsed: {
    cwim: number,
    contrast: number,
  },
  /** 
   * Number of nodes in the original graph (other than the displayed).
   * Due to the limitation of rendering performance, the graph size displayed is much smaller than the actual.
   */
  nNodesRaw: number,
  /**
   * Number of nodes in the original graph (other than the displayed).
   * Due to the limitation of rendering performance, the graph size displayed is much smaller than the actual.
   */
  nLinksRaw: number,
}

/** Display data obtained from JSON input. */
export interface DisplayJSONData 
    extends DisplayDataBase<DisplayLevelJSONData, Array<string>> {}
/** Display data after preprocessing. */
export interface DisplayData 
    extends DisplayDataBase<DisplayLevelData, Set<string>> {}

export interface AnimationConfig {
  /** Maximum FPS of the animation. */
  fps: number,
  /** 
   * # of animating nodes per link during simulation, integer value in range [1, +inf). Denoted as k.
   * Let u -> v be a link with information propagation,
   * there will be k nodes moving from u to v with differing speed in the animation.
   */
  nAnimatingNodesPerLink: number,
  /**
   * Maximum speed of animating nodes, value in range [1, +inf). Denoted as lambda.
   * For each link u -> v during animation, the i-th animation node (i = 1 ... k, k >= 2) requires
   * Fm / Speed(i) frames to reach the destination v, where Speed(i) = (i - 1) / (k - 1) * (lambda - 1) + 1.
   * If k == 1, then Speed(0) = 1.0.
   */
  animatingNodeMaxSpeed: number,
  /** # of frames for moving animation in each section, denoted as Fm. */
  nMovingFrames: number,
  /** # of waiting frames between two moving sections. Denoted as Fw. */
  nWaitingFrames: number,
  styles: {
    nodes: {
      /** Style of seed nodes. */
      seed: NodeStyle,
      /** Style of the visited nodes. */
      visited: NodeStyle,
      /** Style of the animation nodes. */
      animating: NodeStyle,
      /** Style of all other nodes. */
      default?: NodeStyle,
    },
    links: {
      /** Style of the animating links. */
      animating: LinkStyle,
      /** Style of all the other links. */
      default: LinkStyle,
    },
  },
}

export type AnimationPhase = 
    'INITIAL' | 'WAITING' | 'COARSENING' | 'SHOWING_SEEDS' | 'EXPANDING' | 'SIMULATING';

export interface DisplayAnimationStates {
  /** See above */
  config: AnimationConfig,
  /** Whether auto play is enabled. */
  autoStarts: boolean,
  /** Whether manual play is triggered for the next section. */
  manualStarts: boolean,
  /** Current phase of animation. */
  phase: AnimationPhase,
  /** 
   * For coarsening and expanding phase, step = current graph level.
   * For simulating phase, step = current iteration index.
   */
  step: number,
  /** Used in simulating phase. # of iterations whose destination nodes are displayed as visited. */
  nVisitedSteps: number,
  /** Frame index of current animation section. */
  frame: number,
}

export interface ChartsDataItem {
  cwim: number,
  contrast: number
}

export interface ChartsData {
  nNodes: ChartsDataItem,
  timeUsage: ChartsDataItem,
  influence: ChartsDataItem
}

export type ChartsDataItemKey = keyof ChartsDataItem;
export type ChartsDataKey = keyof ChartsData;