import NetV from 'netv';
import * as d3 from 'd3';

import type { 
  LinkData, 
  LinkStyle, 
  NodeData, 
  NodeLinkData, 
  NodeStyle 
} from 'netv/src/interfaces';
import type {
  AnimationConfig,
  DisplayAnimationStates,
  DisplayData, 
  DisplayNodeData,
  IMAlgorithmKey
} from './data_types';
import type Node from 'netv/src/elements/node';

function toStyledNode(node: NodeData, defaultNodeStyle?: NodeStyle): NodeData {
  return {
    ...node,
    style: node.style ?? defaultNodeStyle
  };
}

function toStyledLink(link: LinkData, defaultLinkStyle?: LinkStyle): LinkData {
  return {
    ...link,
    style: link.style ?? defaultLinkStyle
  };
}

function updateNodeStyle(node: Node, style: NodeStyle): void {
  // Common properties
  node.fill(style.fill);
  node.strokeColor(style.strokeColor);
  node.strokeWidth(style.strokeWidth);

  if (node.shape() === 'circle') {
    // (1) Circle: 'r'
    node.r?.call(node, style.r);
  } else if (node.shape() === 'cross' || node.shape() === 'rect') {
    // (2) Rect: 'width' and 'height'
    node.width?.call(node, style.width);
    node.height?.call(node, style.height);
    if (node.shape() === 'cross') {
      // (3) Cross: 'width', 'height', 'innerWidth' and 'innerHeight'
      node.innerWidth?.call(node, style.innerWidth);
      node.innerHeight?.call(node, style.innerHeight);
    }
  } else if (node.shape() === 'triangle') {
    // (4) Triangle: 'vertexAlpha', 'vertexBeta' and 'vertexGamma'
    node.vertexAlpha?.call(node, style.vertexAlpha);
    node.vertexBeta?.call(node, style.vertexBeta);
    node.vertexGamma?.call(node, style.vertexGamma);
  } else {
    throw TypeError(`Invalid node shape '${node.shape()}'.`);
  }
}

function nodeLinkLimitOfGraphCanvas(data: DisplayData) {
  const n = data.graphs[0].nodes.size;
  const m = data.graphs[0].links.length;
  return {n: n, m: m};
}

function nodeLinkLimitOfAnimatingCanvas(data: DisplayData, config: AnimationConfig) {
  const maxNLinks = Math.max(
    d3.max(data.simulation.cwim.map((links) => links.length)) as number,
    d3.max(data.simulation.contrast.map((links) => links.length)) as number
  );
  const n = data.graphs[0].nodes.size + config.nAnimatingNodesPerLink * maxNLinks;
  const m = maxNLinks;
  return {n: n, m: m};
}

export function nodeLinkLimitOfCanvases(data: DisplayData, config: AnimationConfig) {
  return {
    graphCanvas: nodeLinkLimitOfGraphCanvas(data),
    animatingCanvas: nodeLinkLimitOfAnimatingCanvas(data, config)
  };
}

export function createEmptyCanvas(div: HTMLDivElement, nodeLimit: number, linkLimit: number): NetV;
export function createEmptyCanvas(div: HTMLDivElement, limit: {n: number, m: number}): NetV;

export function createEmptyCanvas(
    div: HTMLDivElement,
    nodeLimit: number | {n: number, m: number},
    linkLimit?: number
): NetV {
  const n = (typeof nodeLimit === 'number') ? nodeLimit : nodeLimit.n;
  const m = (typeof nodeLimit === 'number') ? linkLimit : nodeLimit.m;
  const netv = new NetV({
    container: div,
    width: div.offsetWidth,
    height: div.offsetHeight,
    nodeLimit: n,
    linkLimit: m,
    backgroundColor: { r: 0, g: 0, b: 0, a: 0 } // Transparent
  });
  const L = Math.min(div.offsetWidth, div.offsetHeight);
  const D = div.offsetWidth - div.offsetHeight;
  if (D >= 0) {
    // Coordinate range is transformed linearly from [0, 1], [0, 1] to [D/2, L + D/2], [0, L].
    netv.transform({x: D / 2, y: 0, k: L});
  } else {
    // Coordinate range is transformed linearly from [0, 1], [0, 1] to [0, L], [L - D/2, -D/2].
    netv.transform({x: 0, y: -D / 2, k: L});
  }
  return netv;
}

export function drawInitialCanvas(
    netv: NetV, 
    rawNodes?: Array<NodeData> | Set<NodeData> | Map<string, NodeData>, 
    rawLinks?: Array<LinkData>, 
    defaultNodeStyle?: NodeStyle, 
    defaultLinkStyle?: LinkStyle
): void {
  const nodes: Array<NodeData> = []
  if (rawNodes !== null && rawNodes !== undefined) {
    if (rawNodes instanceof Array || rawNodes instanceof Set) {
      rawNodes.forEach((node) => {
        nodes.push(toStyledNode(node, defaultNodeStyle));
      });
    } else if (rawNodes instanceof Map) {
      rawNodes.forEach((node, _) => {
        nodes.push(toStyledNode(node, defaultNodeStyle));
      });
    } else {
      throw TypeError("Expects rawNodes to be either one of Array, Set or Map.");
    } 
  }
  const linksFn = () => {
    if (rawLinks !== null && rawLinks !== undefined) {
      return Array.from(rawLinks, (link) => toStyledLink(link, defaultLinkStyle));
    }
    return [];
  }
  const links = linksFn();

  netv.data({ nodes: nodes, links: links });
  netv.draw();
}

export function drawInitialGraphCanvas(
    graphCanvas: NetV,
    data: DisplayData,
    config: AnimationConfig
): void {
  drawInitialCanvas(
    graphCanvas, 
    data.graphs[0].nodes, 
    data.graphs[0].links, 
    config.styles.nodes.default, 
    config.styles.links.default,
  );
}

export function initialCoarseningAnimationStates(
    config: AnimationConfig
): DisplayAnimationStates {
  return {
    config: config,
    autoStarts: false,
    manualStarts: false,
    phase: 'INITIAL',
    step: 0,
    nVisitedSteps: 0,
    frame: 0
  };
}

function proceedFrameIndex(
    states: DisplayAnimationStates,
    nFrames: number,
    nSteps?: number,
): boolean {
  if (nSteps && states.step < 0) {
    states.step = 0;
    states.frame = 0;
  }
  // Proceeding
  states.frame += 1;
  if (states.frame >= nFrames) {
    states.frame = 0;
    if (nSteps) {
      states.step += 1;
      if (states.step >= nSteps) {
        states.step = 0;
        return false; // Stops
      }
    } else {
      return false;
    }
  }
  return true; // Continues
}

function proceedNoOpAnimationFrame(
    states: DisplayAnimationStates,
    nFrames: number,
    nSteps?: number,
): boolean {
  return proceedFrameIndex(states, nFrames, nSteps);
}

function proceedCoarseningAnimationFrame(
    netv: NetV, 
    data: DisplayData, 
    states: DisplayAnimationStates
): boolean {
  if (states.step < 0) {
    states.step = 0;
    states.frame = 0;
  }
  // Stops if the last layer is reached.
  if (states.step >= data.graphs.length) {
    return false;
  }
  const nTotalFrames = states.config.nMovingFrames + states.config.nWaitingFrames;
  
  if (states.frame === 0) {
    // (1) The first frame of new level
    const rawNodes = Array.from(
      data.graphs[states.step].nodes.values()
    );
    const rawLinks = data.graphs[states.step].links;
    const styles = states.config.styles;
    netv.data({
      nodes: rawNodes.map(
        (node) => toStyledNode(node, styles.nodes.default)
      ),
      links: rawLinks.map(
        (link) => toStyledLink(link, styles.links.default)
      )
    });
  } else if (states.step < data.graphs.length - 1 && states.frame < states.config.nMovingFrames) {
    // (2) The next frame of current level
    const curNodeMap = data.graphs[states.step].nodes;
    const nextNodeMap = data.graphs[states.step + 1].nodes;

    netv.nodes().forEach((node: any) => {
      const from = curNodeMap.get(node.id());
      if (!from) {
        throw RangeError(`Error with node ID '${node.id()}.`);
      }
      const to = nextNodeMap.get(from.to as string);
      if (!to) {
        throw RangeError(`Error with node ID ${from.to}.`);
      }
      const lambda = states.frame / (states.config.nMovingFrames - 1);
      node.x(from.x + lambda * (to.x - from.x));
      node.y(from.y + lambda * (to.y - from.y));
    });
  } // Otherwise, nothing to todo
  // Refreshes the canvas
  netv.draw();
  // Proceeding
  states.frame += 1;
  if (states.frame == nTotalFrames) {
    states.frame = 0;
    states.step += 1;
  }
  return (states.step < data.graphs.length);
}

function proceedShowingSeedsFrame(
    netv: NetV,
    seeds: Array<string> | Set<string>,
    states: DisplayAnimationStates,
): boolean {
  const nTotalFrames = states.config.nMovingFrames + states.config.nWaitingFrames;
  const seedStyle = states.config.styles.nodes.seed;
  if (states.frame === 0) {
    seeds.forEach((s) => {
      const node = netv.getNodeById(s);
      if (node) {
        updateNodeStyle(node, seedStyle);
      } else {
        console.warn(`Seed displaying error: Seed ${s} is missing.`);
      }
    });
    netv.draw(); // Refreshes the canvas
  }
  // Proceeding
  states.frame += 1;
  if (states.frame >= nTotalFrames) {
    states.frame = 0;
    return false; // Stops
  }
  return true; // Continues
}

function proceedExpandingAnimationFrame(
    netv: NetV, 
    data: DisplayData, 
    states: DisplayAnimationStates
): boolean {
  if (states.step > data.graphs.length - 1) {
    states.step = data.graphs.length - 1;
    states.frame = 0;
  }
  // Stops if the last layer is reached.
  if (states.step < 0) {
    return false;
  }
  const nTotalFrames = states.config.nMovingFrames + states.config.nWaitingFrames;
  const styles = states.config.styles;

  if (states.frame === 0) {
    // (1) The first frame of new level
    const curNodeMap = data.graphs[states.step].nodes;
    const nonSeedStyle = (node: NodeData) => node.style ?? styles.nodes.default;
    if (states.step > 0) {
      // (1.1) Not the first level
      const nodes: Array<DisplayNodeData> = [];
      data.graphs[states.step - 1].nodes.forEach((node, id) => {
        const from = curNodeMap.get(node.to as string);
        if (from === undefined) {
          throw Error(`Error with node ID '${node.to}'.`);
        }
        const style = data.seeds.cwim[states.step - 1].has(id) ? styles.nodes.seed : nonSeedStyle(node);
        nodes.push({id: id, x: from.x, y: from.y, style: style});
      });
      netv.data({
        nodes: nodes,
        links: data.graphs[states.step - 1].links.map(
          (link) => toStyledLink(link, styles.links.default)
        )
      });
    } else {
      // (1.2) The first level
      const nodes: Array<DisplayNodeData> = [];
      data.graphs[0].nodes.forEach((node, id) => {
        const style = data.seeds.cwim[0].has(id) ? styles.nodes.seed : nonSeedStyle(node);
        nodes.push({id: id, x: node.x, y: node.y, style: style});
      });
      netv.data({
        nodes: nodes,
        links: data.graphs[0].links.map(
          (link) => toStyledLink(link, styles.links.default)
        )
      });
    }
  } else if (states.step > 0 && states.frame < states.config.nMovingFrames) {
    // (2) The next frame of current level
    const prevNodeMap = data.graphs[states.step - 1].nodes;
    netv.nodes().forEach((node: any) => {
      const from = prevNodeMap.get(node.id());
      if (!from) {
        throw Error(`Error with ID '${node.id()}.`);
      }
      const lambda = states.frame / (states.config.nMovingFrames - 1);
      const x = node.x();
      const y = node.y();
      node.x(x + lambda * (from.x - x));
      node.y(y + lambda * (from.y - y));
    });
  } // Otherwise, nothing to todo
  // Refreshes the canvas
  netv.draw(); 
  // Proceeding
  states.frame += 1;
  if (states.frame == nTotalFrames) {
    states.frame = 0;
    states.step -= 1;
  }
  return (states.step >= 0);
}

function resetGraphCanvas(graphCanvas: NetV, data: DisplayData, config: AnimationConfig): void {
  drawInitialGraphCanvas(graphCanvas, data, config);
}

function resetAnimatingCanvas(animatingCanvas: NetV): void {
  animatingCanvas.data({ nodes: [], links: [] });
  animatingCanvas.draw(); // Resets with empty data
}

function updateGraphCanvasByVisited(
    graphCanvas: NetV,
    toStep: number,
    simData: Array<Array<LinkData>>,
    states: DisplayAnimationStates,
): void {
  if (states.nVisitedSteps >= toStep) {
    return; // No-op
  }
  const visitedStyle = states.config.styles.nodes.visited;

  for (; states.nVisitedSteps < toStep; ++states.nVisitedSteps) {
    simData[states.nVisitedSteps].forEach((link) => {
      const node = graphCanvas.getNodeById(link.target);
      if (node) {
        updateNodeStyle(node, visitedStyle);
      } else {
        console.warn(`Missing node with ID '${link.target}'.`);
      }
    });
  }
  graphCanvas.draw();
}

function proceedSimulatingAnimationFrame(
    graphCanvas: NetV,
    animatingCanvas: NetV,
    data: DisplayData,
    states: DisplayAnimationStates,
    algorithmKey: IMAlgorithmKey,
): boolean {
  const config = states.config;
  const simData = data.simulation[algorithmKey];
  const maxNSteps = d3.max(Object.entries(data.simulation).map(([_, arr]) => arr.length)) as number;
  const nTotalFrames = config.nMovingFrames + config.nWaitingFrames;

  if (simData.length === 0) {
    console.warn("linksByStep is empty. Nothing for animation.");
    return false; // Stops right now
  }
  // One extra step for resetting
  if (states.step >= maxNSteps) {
    if (states.frame === 0) {
      resetGraphCanvas(graphCanvas, data, states.config);
      resetAnimatingCanvas(animatingCanvas);
    }
    states.frame += 1;
    if (states.frame >= nTotalFrames) {
      states.step = states.nVisitedSteps = states.frame = 0;
      return false; // Finishes
    }
    return true; // Continues
  }
  // Waiting for (1) other canvas; (2) animation pause interval
  if (states.step >= simData.length || states.frame >= config.nMovingFrames) {
    states.frame += 1;
    if (states.frame >= nTotalFrames) {
      states.step += 1;
      states.frame = 0;
    }
    return true;
  }
  const lambda = states.frame / (config.nMovingFrames - 1);
  const animationData: NodeLinkData = { nodes: [], links: [] };

  // Animation Links
  const animatingLinkStyle = config.styles.links.animating;
  if (animatingLinkStyle.strokeColor) {
    const strokeColor = animatingLinkStyle.strokeColor;
    const linkStyle = {
      strokeColor: {
        r: strokeColor.r,
        g: strokeColor.g,
        b: strokeColor.b,
        a: strokeColor.a * (1.0 - lambda)
      },
      strokeWidth: animatingLinkStyle.strokeWidth
    };
    if (states.frame === 0) {
      animationData.links = simData[states.step].map((link) => {
        return { source: link.source, target: link.target, style: linkStyle };
      });
    } else {
      animatingCanvas.links().forEach((link: any) => {
        link.strokeColor(linkStyle.strokeColor); // Modified intrusively
      });
    }
  }

  // Animation nodes
  const pastRatioFn = () => {
    if (config.nAnimatingNodesPerLink <= 1) {
      return [1.0];
    }
    return d3.range(config.nAnimatingNodesPerLink).map((j) => {
      const speed = 1.0 + j * (config.animatingNodeMaxSpeed - 1.0) / (config.nAnimatingNodesPerLink - 1);
      return Math.min(1.0, speed * lambda);
    });
  };
  const pastRatio = pastRatioFn();
  // The first animation node reaches the destination.
  if (pastRatio[pastRatio.length - 1] >= 1.0) {
    updateGraphCanvasByVisited(graphCanvas, states.step + 1, simData, states);
  }

  const currentAnimatingLocationsOf = (link: LinkData) => {
    const source = graphCanvas.getNodeById(link.source);
    if (!source) {
      throw Error(`Invalid link source '${link.source}'.`);
    }
    const sx = source.x();
    const sy = source.y();
    const target = graphCanvas.getNodeById(link.target);
    if (!target) {
      throw Error(`Invalid link target '${link.target}.`);
    }
    const tx = target.x();
    const ty = target.y();
    return pastRatio.map((r) => [sx + r * (tx - sx), sy + r * (ty - sy)]);
  };

  if (states.frame === 0) {
    // (1) Animating nodes
    simData[states.step].forEach((link: LinkData, i: number) => {
      currentAnimatingLocationsOf(link).forEach(([x, y], j) => {
        const newNode = { id: `$animating-${i}-${j}`, x: x, y: y, style: config.styles.nodes.animating };
        animationData.nodes.push(newNode);
      });
    });
    // (2) Link ends (Invisible)
    const involvedNodeIDs = new Set<string>();
    simData[states.step].forEach(({source, target}) => {
      involvedNodeIDs.add(source);
      involvedNodeIDs.add(target);
    });
    involvedNodeIDs.forEach((id) => {
      const node = graphCanvas.getNodeById(id);
      if (!node) {
        throw Error(`Invalid node ID '${id}'.`);
      }
      const x = node.x();
      const y = node.y();
      const invisibleStyle = { r: 0, strokeWidth: 0, fill: { r: 0, g: 0, b: 0, a: 0 } };
      animationData.nodes.push({ id: id, x: x, y: y, style: invisibleStyle });
    });
  } else {
    // Updates locations of animating nodes intrusively
    const nodes = animatingCanvas.nodes();
    simData[states.step].forEach((link, i) => {
      currentAnimatingLocationsOf(link).forEach(([x, y], j) => {
        const cur = nodes[i * config.nAnimatingNodesPerLink + j];
        cur.x(x);
        cur.y(y); // Updates (x, y) intrusively
      });
    });
  }

  if (states.frame === 0) {
    animatingCanvas.data(animationData);
  }
  animatingCanvas.draw();

  // Proceeding
  states.frame += 1;
  if (states.frame >= nTotalFrames) {
    states.step += 1;
    states.frame = 0;
  }
  return true;
}

export function proceedCWIMAnimationFrame(
    graphCanvas: NetV,
    animatingCanvas: NetV,
    data: DisplayData,
    states: DisplayAnimationStates
): void {
  const doesProceed = states.autoStarts || states.manualStarts || states.frame > 0;
  if (!doesProceed) {
    return; // No-op
  }
  states.manualStarts = false;

  if (states.phase === 'INITIAL') {
    states.phase = 'COARSENING';
  }
  // Animation phase loop
  if (states.phase === 'COARSENING') {
    if (!proceedCoarseningAnimationFrame(graphCanvas, data, states)) {
      states.phase = 'SHOWING_SEEDS';
      console.log("COARSENING => SHOWING_SEEDS");
    }
  } else if (states.phase === 'SHOWING_SEEDS') {
    const seeds = data.seeds.cwim[data.seeds.cwim.length - 1];
    const continues = proceedShowingSeedsFrame(graphCanvas, seeds, states);
    if (!continues) {
      states.phase = 'WAITING';
      console.log("SHOWING_SEEDS => WAITING");
    }
  } else if (states.phase === 'WAITING') {
    const nFrames = states.config.nMovingFrames + states.config.nWaitingFrames;
    const continues = proceedNoOpAnimationFrame(states, nFrames);
    if (!continues) {
      states.phase = 'EXPANDING';
      console.log("WAITING => EXPANDING");
    }
  } else if (states.phase === 'EXPANDING') {
    if (!proceedExpandingAnimationFrame(graphCanvas, data, states)) {
      states.phase = 'SIMULATING';
      states.step = states.nVisitedSteps = states.frame = 0;
      console.log("EXPANDING => SIMULATING");
    }
  } else if (states.phase === 'SIMULATING') {
    if (!proceedSimulatingAnimationFrame(graphCanvas, animatingCanvas, data, states, 'cwim')) {
      states.phase = 'COARSENING';
      console.log("SIMULATING => COARSENING");
    }
  } else {
    throw TypeError(`Invalid animation phase '${states.phase}'.`);
  }
}

export function proceedContrastAnimationFrame(
    graphCanvas: NetV,
    animatingCanvas: NetV,
    data: DisplayData,
    states: DisplayAnimationStates
): void {
  const doesProceed = states.autoStarts || states.manualStarts || states.frame > 0;
  if (!doesProceed) {
    return; // No-op
  }
  states.manualStarts = false;

  if (states.phase === 'INITIAL') {
    states.phase = 'WAITING';
  }
  // Animation phase loop
  const nFrames = states.config.nMovingFrames + states.config.nWaitingFrames;
  if (states.phase === 'WAITING') {
    if (states.frame === 0 && states.step === 0) {
      drawInitialGraphCanvas(graphCanvas, data, states.config);
    }
    const nLevels = data.graphs.length * 2 + 1;
    if (!proceedNoOpAnimationFrame(states, nFrames, nLevels)) {
      states.phase = 'SHOWING_SEEDS';
      console.log("Contrast: WAITING (no-op) => SHOWING_SEEDS");
    }
  } else if (states.phase === 'SHOWING_SEEDS') {
    const continues = proceedShowingSeedsFrame(graphCanvas, data.seeds.contrast, states);
    if (!continues) {
      states.phase = 'SIMULATING';
      console.log("Contrast: SHOWING_SEEDS => SIMULATING");
    }
  } else if (states.phase === 'SIMULATING') {
    if (!proceedSimulatingAnimationFrame(graphCanvas, animatingCanvas, data, states, 'contrast')) {
      states.phase = 'WAITING';
      console.log("Contrast: SIMULATING => WAITING (no-op)");
    }
  } else {
    throw TypeError(`Invalid animation phase '${states.phase}'.`);
  }
}

export function getProceedAnimationFrameFn(algorithmKey: string) {
  if (algorithmKey === 'cwim') {
    return proceedCWIMAnimationFrame;
  } else if (algorithmKey === 'contrast') {
    return proceedContrastAnimationFrame;
  } else {
    throw RangeError(`Invalid algorithm key '${algorithmKey}'.`);
  }
}
